/*
 * Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
 *
 * This software is dual-licensed to you under the MIT License (MIT) and
 * the Universal Permissive License (UPL).  See the LICENSE file in the root
 * directory for license terms.  You may choose either license, or both.
 */

package com.oracle.iot.client;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;

/**
 * An unexpected HTTP response from the server is thrown as an TransportException.
 * This class extends IOException so that exceptions of this type can be
 * thrown by client library API.
 */
public class TransportException extends IOException {

    /**
     * Create an TransportException with an HTTP status code.
     * @param statusCode an <a href="https://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html">HTTP status code</a>}
     */
    public TransportException(int statusCode) {
        super();
        this.statusCode = statusCode;
        this.data = null;
        this.headers = null;
    }

    /**
     * Create an TransportException with a status code and a message.
     * @param statusCode an <a href="https://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html">HTTP status code</a>}
     * @param message additional information that may be useful for diagnostics.
     */
    public TransportException(int statusCode, String message) {
        super(message);
        this.statusCode = statusCode;
        this.data = null;
        this.headers = null;
    }

    public TransportException(String method, String path, HttpResponse httpResponse) {
        super(httpResponse.getVerboseStatus(method, path));
        this.statusCode = httpResponse.getStatus();

        try {
            final byte[] responseData = httpResponse.getData();
            this.data = responseData != null ? new String(responseData, "UTF-8") : null;
        } catch (UnsupportedEncodingException cannot_happen) {
            // this cannot happen because UTF-8 is a required encoding
            // throw RuntimeException to satisfy the compiler
            throw new RuntimeException(cannot_happen);
        }

        final Map<String, List<String>> responseHeaders = httpResponse.getHeaders();
        if (responseHeaders != null) {
            // Normalize the header keys by forcing them all to lower case.
            // This helps simplify the lookup code.
            headers = new HashMap<String, List<String>>(responseHeaders.size());
            for (Map.Entry<String,List<String>> entry : responseHeaders.entrySet()) {
                headers.put(
                    (entry.getKey() != null ? entry.getKey().toLowerCase(Locale.ROOT) : null),
                    entry.getValue()
                );
            }
        } else {
            headers = null;
        }
    }

    /**
     * Get HTTP status code.
     * @return an <a href="https://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html">HTTP status code</a>}
     */
    public int getStatusCode() {
        return statusCode;
    }

    /**
     * Get whatever data may have been returned by the server. For an HTTP connection, this
     * is the data from the error stream, which may be empty.
     * @return data from the error, or {@code null} if there is no such data.
     */
    public String getResponseData() {
        return data;
    }

    /**
     * Get the value of a header which may have been returned by the server. For an HTTP
     * connection, these are the headers in the HTTP response. If there is more than one
     * value in the response header, the first value is returned. The {@code header}
     * parameter is not case sensitive.
     * @param header the name of the header, e.g., "Retry-After".
     * @return the first value of header, or {@code null}
     */
    public String getResponseHeader(String header) {
        if (headers != null) {
            final List<String> values = headers.get(header.toLowerCase(Locale.ROOT));
            if (values != null && values.size() > 0) {
                return values.get(0);
            }
        }
        return null;
    }

    @Override
    public String toString() {
        final String message = "HTTP " + statusCode;
        if (getMessage() != null) {
            return message + ": " + getMessage();
        }
        return message;
    }

    private final int statusCode;
    private final String data;
    private final Map<String, List<String>> headers;
}
