/*
 * Copyright (c) 2017, 2019, Oracle and/or its affiliates.  All rights reserved.
 *
 * This software is dual-licensed to you under the MIT License (MIT) and 
 * the Universal Permissive License (UPL).  See the LICENSE file in the root
 * directory for license terms.  You may choose either license, or both.
 */

package com.oracle.iot.client.impl;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import org.json.JSONException;
import org.json.JSONObject;

/**
 * StorageAuthenticationResponse
 */
class StorageAuthenticationResponse {

    private static final String FIELD_STORAGE_CONTAINER_URL  = "storageContainerUrl";
    private static final String FIELD_AUTH_TOKEN    = "authToken";

    private String storageContainerUrl;
    private String authToken;
    private Map<String,String> headers;

    private StorageAuthenticationResponse() {}

    String getAuthToken() {
        return authToken;
    }

    String getStorageContainerUrl() {
        return storageContainerUrl;
    }

    Map<String,String> getHeaders() { return headers; }

    @Override
    public String toString() {
        return "StorageAuthenticationResponse{" +
                "storageContainerUrl='" + storageContainerUrl + '\'' +
                ", authToken=" + authToken +
                ", headers=" + headers +
                '}';
    }

    public static StorageAuthenticationResponse fromJson(final JSONObject 
            jsonObject) {
        StorageAuthenticationResponse response = new StorageAuthenticationResponse();

        //
        // For Object Storage, response looks like
        // {
        //   "storageUrl": "https://{server}/v1/c/{compartment}/b/{bucket}/o/{object}",
        //   "headers: {
        //        "Authorization": "Signature keyId=\"ocid1.tenan...,algorithm=...,headers=...,signature=...\"",
        //        "date":"some date in SimpleDateFormat: EEE, dd MMM yyyy HH:mm:ss zzz",
        //        "host":"URL of host",
        //        "(request-target)": "get /v1/iotdev/Test"
        //   }
        // }
        //
        // For Object Storage Classic, response looks like
        // {
        //    "storageContainerUrl":"https://{instance}/v1/{account}/{container}",
        //    "authToken":"AUTH_tk9e18bb3eba2eb0c5d16847218a12a36f"
        // }
        //
        try {
            if (jsonObject.has(FIELD_STORAGE_CONTAINER_URL)) {
                response.storageContainerUrl = jsonObject.getString(FIELD_STORAGE_CONTAINER_URL);
            } else {
                response.storageContainerUrl = jsonObject.getString("storageUrl");
            }

            if (jsonObject.has(FIELD_AUTH_TOKEN)) {
                response.authToken = jsonObject.getString(FIELD_AUTH_TOKEN);
            } else {
                response.headers = new HashMap<String,String>();
                JSONObject headers = jsonObject.getJSONObject("headers");
                Iterator<String> keys = headers.keys();
                while (keys.hasNext()) {
                    String key = keys.next();
                    String value = headers.getString(key);
                    response.headers.put(key, value);
                }
                response.authToken = response.headers.get("Authorization");
            }

        } catch (JSONException e) {
            throw new RuntimeException(e);
        }

        return response;
    }
    
    public String toJson() {
        JSONObject objectBuilder = new JSONObject();

        try {
            objectBuilder.put(FIELD_STORAGE_CONTAINER_URL, storageContainerUrl);
            objectBuilder.put(FIELD_AUTH_TOKEN, authToken);
        } catch (JSONException e) {
            throw new RuntimeException(e);
        }

        return objectBuilder.toString();
    }
}
