/*
 * Copyright (c) 2017, 2019, Oracle and/or its affiliates.  All rights reserved.
 *
 * This software is dual-licensed to you under the MIT License (MIT) and 
 * the Universal Permissive License (UPL).  See the LICENSE file in the root
 * directory for license terms.  You may choose either license, or both.
 */
package com.oracle.iot.client.impl;

import com.oracle.iot.client.TransportException;
import com.oracle.iot.client.HttpResponse;
import com.oracle.iot.client.RestApi;
import com.oracle.iot.client.SecureConnection;

import com.oracle.iot.client.impl.util.Base64;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.channels.FileChannel;
import java.security.GeneralSecurityException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.HashMap;
import java.util.Locale;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.Map;
import java.util.List;
import java.util.Set;

import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.json.JSONException;
import org.json.JSONObject;
import com.oracle.iot.client.StorageObject;

/**
 * The StorageConnection transfers content to the Storage Cloud Service.
 * Service. There is one StorageConnection instance per client.
 */
public abstract class StorageConnectionBase implements StorageConnection {

    //
    // Oracle storage cloud host pattern matches URIs that start with
    //     https://objectstorage.*.oraclecloud.com
    // or
    //     https://*.storage.oraclecloud.com
    //
    // An object name in the storage cloud matches URIs that start with
    //     https://objectstorage.*.oraclecloud.com/n/*/b/*/o/{object}
    //
    // An object name in the classic storage could matches URIs that start with
    //     https://*.storage.oraclecloud.com/v1/*/*/{object}
    //
    // Note that patterns make assumptions that the URI string being matched is valid.
    // Note that the port number on the URL is treated as optional.
    // Note that the 'localhost' optional match is for mock server.
    // Note that the use of [^/] is to match any character except the forward slash. This
    // ensures that the object name is correctly grouped for a classic storage URI
    // of the form https://*.storage.oraclecloud.com/v1/*/*/{endpoint-id}/{file-name}. If
    // the pattern was .+ (any character), the match would group just the {file-name}.
    // Alternatively, the pattern .+? (non-greedy match any) could have been used, but
    // 'match any character except forward slash' is easier to understand.
    //
    private static final Pattern OBJECT_STORAGE_HOST_PATTERN =
        Pattern.compile("https?://(?:(?:objectstorage\\..+|.+\\.storage)\\.oraclecloud\\.com|localhost).*");

    private static final Pattern OBJECT_STORAGE_PATTERN =
        Pattern.compile("https?://(?:objectstorage\\..+\\.oraclecloud\\.com|localhost)(?::\\d+)?/n/[^/]+/b/[^/]+/o/(.+)");

    private static final Pattern OBJECT_STORAGE_CLASSIC_PATTERN =
        Pattern.compile("https?://(?:.+\\.storage\\.oraclecloud\\.com|localhost)(?::\\d+)?/v1/[^/]+/[^/]+/(.+)");

    private static final boolean virtualStorageDirectories;
    /**
     * REST resource for storage authentication
     */
    private static final String REST_STORAGE_AUTHENTICATION =
        RestApi.V2.getReqRoot()+"/provisioner/storage";
    private static final String AUTH_TOKEN_HEADER = "X-Auth-Token";
    private static final int DEFAULT_RETRY_LIMIT = 2;
    private static final int ERR_BAD_CHECKSUM = -1;
    private static final int ERR_UPLOAD_CANCELLED = -2;
    private static final int ERR_WRITING_REQUEST_BODY_TO_SERVER = -3;
    /**
     * Number of times to retry a transfer if the token expires.
     */
    private static final int RETRY_LIMIT;
    /**
     * Default chunk size is 4096.
     */
    private static final int DEFAULT_CHUNK_SIZE = 4096;
    /**
     * Property for setting chunk size. No limits imposed on chunk size,
     * but HttpURLConnection may impose its own limits.
     */
    private static final String CHUNK_SIZE_PROPERTY = "com.oracle.iot.client.storage_connection_chunk_size";
    /**
     * Size of chunks for chunked streaming mode.
     */
    private static final int CHUNK_SIZE = Integer.getInteger(CHUNK_SIZE_PROPERTY, DEFAULT_CHUNK_SIZE);

    /**
     * If using Object Storage, we pass a JSON body to the POST /api/v2/provisioner/storage API
     * to get an authorization header. For Object Storage Classic, there is no body to the
     * GET /api/v2/provisioner/storage API. When Object Storage Classic API is called for,
     * AUTHENTICATION_REQUEST_BODY_IS_NOT_NEEDED is passed to the setStorageAuthenticationData
     * method.
     */
    private static final byte[] AUTHENTICATION_REQUEST_BODY_IS_NOT_NEEDED = null;

    private final SecureConnection secureConnection;
    private StorageAuthenticationResponse  storageAuthenticationData = null;

    private boolean closed;

    static {
        Integer val = Integer.getInteger("com.oracle.iot.client.storage_connection_retry_limit");
        RETRY_LIMIT = ((val != null) && (val > 0))? val : DEFAULT_RETRY_LIMIT;

        final String value = System.getProperty("oracle.iot.client.disable_storage_object_prefix");
        virtualStorageDirectories = value == null || "".equals(value) || !Boolean.parseBoolean(value);
    }

    protected StorageConnectionBase(SecureConnection secureConnection) {
        this.secureConnection = secureConnection;
        this.closed = false;
    }

    @Override
    final public void sync(StorageObject storageObject) throws IOException, GeneralSecurityException {
        if (storageObject.getInputStream() != null) {
            upload(storageObject);
        }
        else if (storageObject.getOutputStream() != null) {
            download(storageObject);
        }
        else throw new IllegalArgumentException("InputStream and OutputStream are not set.");
    }

    private void upload(StorageObject storageObject) throws IOException, GeneralSecurityException {
        int responseCode = transfer(storageObject);
        if (responseCode != HttpURLConnection.HTTP_CREATED && responseCode != HttpURLConnection.HTTP_OK) {
            throw new TransportException(responseCode, "Upload " + storageObject.getURI());
        }
    }

    private void download(StorageObject storageObject) throws IOException, GeneralSecurityException {
        int responseCode = transfer(storageObject);
        if (responseCode != HttpURLConnection.HTTP_OK && responseCode != HttpURLConnection.HTTP_PARTIAL) {
            throw new TransportException(responseCode, "Download " + storageObject.getURI());
        }
    }

    final public StorageObject createStorageObject(String clientId,
            String name, String contentType)
            throws IOException, GeneralSecurityException {

        final String objectName;
        if (virtualStorageDirectories && clientId != null) {
            objectName = clientId + "/" + name;
        } else {
            objectName = name;
        }

        // In order to create the StorageObject, we need to form the URI.
        // To form the URI, we need to get the storage authentication data.
        // But we don't know whether the storage service configured on IoT CS
        // is for Object Storage or Object Storage Classic. We could look at the current
        // storage authentication data (if we have it), but there is a (slim)
        // chance that the storage service configuration on IoT CS has changed.
        // Therefore, we always get the storage authentication data when
        // creating the storage object since it will give the correct URI.
        // Since we don't know which API to use, we assume Object Storage
        // (POST /api/v2/provisioner/storage) and fall back to Object Storage Classic
        // (GET /api/v2/provisioner/storage). This should be most often correct
        // with post 19.1.1 versions of the IoT CS.

        boolean isClassicObjectStorage = false;
        try {
            final byte[] authenticationRequestBody
                = createAuthenticationRequestBody(objectName, "HEAD");
            setStorageAuthenticationData(authenticationRequestBody);
        } catch (TransportException e) {
            if (e.getStatusCode() == HttpURLConnection.HTTP_PRECON_FAILED
                || e.getStatusCode() == HttpURLConnection.HTTP_BAD_METHOD) {
                isClassicObjectStorage = true;
                setStorageAuthenticationData(AUTHENTICATION_REQUEST_BODY_IS_NOT_NEEDED);
            } else {
                throw e;
            }
        }

        // Classic storage API gives us the URI of the container, to which we have to append
        // the objectName. Object Storage gives us the full URL.
        final String contentStorageLocation;
        if (isClassicObjectStorage) {
            contentStorageLocation = storageAuthenticationData.getStorageContainerUrl() + "/" + objectName;
        } else {
            contentStorageLocation = storageAuthenticationData.getStorageContainerUrl();
        }
        final StorageObject storageObject =
            createStorageObject(contentStorageLocation, name, // name is correct here!
                                contentType, null, null, -1);

        return storageObject;
    }

    final public StorageObject createStorageObject(String storageUrl)
        throws IOException, GeneralSecurityException {

        final URL url;
        try {
            url = new URL(storageUrl);
        } catch (MalformedURLException ex) {
            throw new IllegalArgumentException("Storage Cloud URL is invalid", ex);
        }

        final String name = getObjectNameFromURL(url);
        final boolean isClassicObjectStorage = pathIsClassicObjectStore(url.getPath());
        final byte[] authenticationRequestBody
            = isClassicObjectStorage
            ? AUTHENTICATION_REQUEST_BODY_IS_NOT_NEEDED
            : createAuthenticationRequestBody(name, "HEAD");

        setStorageAuthenticationData(authenticationRequestBody);

        if (!storageUrl.startsWith(storageAuthenticationData.getStorageContainerUrl())) {
            throw new GeneralSecurityException("Storage container URL does not match.");
        }

        int i=0;
        while (i++ < RETRY_LIMIT) {
            final HttpURLConnection connection = (HttpURLConnection) url.openConnection();
            connection.setRequestMethod("HEAD");
            if (isClassicObjectStorage) {
                connection.setRequestProperty(AUTH_TOKEN_HEADER, storageAuthenticationData.getAuthToken());
            } else {
                Map <String,String> headers = storageAuthenticationData.getHeaders();
                for(Map.Entry<String,String> header : headers.entrySet()) {
                    connection.setRequestProperty(header.getKey(), header.getValue());
                }
            }
            connection.connect();

            final int length = connection.getContentLength();
            final String type = connection.getContentType();
            final String encoding = connection.getContentEncoding();
            final String date = connection.getHeaderField("Last-Modified");
            final int responseCode = connection.getResponseCode();

            if (responseCode == HttpURLConnection.HTTP_OK) {
                return createStorageObject(storageUrl, name, type, encoding, date, length);
            } else if (responseCode != HttpURLConnection.HTTP_UNAUTHORIZED) {
                throw new TransportException(responseCode, "HEAD " + storageUrl);
            } else {
                storageAuthenticationData = null;
                if (i < RETRY_LIMIT) {
                    setStorageAuthenticationData(authenticationRequestBody);
                }
            }
        }
        throw new TransportException(HttpURLConnection.HTTP_UNAUTHORIZED, "HEAD " + storageUrl);
    }

    protected abstract StorageObject createStorageObject(
                                           String storageContainerUrl,
                                           String name,
                                           String type,
                                           String encoding,
                                           String date,
                                           int length);

    public static boolean isStorageCloudURI(String stringUri) {

        try {
            final URI uri = new URI(stringUri);
            final Matcher matcher = OBJECT_STORAGE_HOST_PATTERN.matcher(uri.toString());
            return matcher.matches();
        } catch (URISyntaxException e) {
            // ignored
        }
        return false;
    }

    @Override
    final public synchronized void close() throws IOException {
        if (!closed) {
            closed = true;
        }
    }

    private static int getContentLength(InputStream inputStream) throws IOException {

        if (inputStream.markSupported() || (inputStream instanceof FileInputStream)) {

            if (inputStream.markSupported()) {
                inputStream.mark(Integer.MAX_VALUE);
            }

            long total = 0;
            try {
                int len;
                final byte[] buf = new byte[4096];
                while ((len = inputStream.read(buf)) != -1) {
                    total += len;
                }
            }  finally {
                if (inputStream.markSupported()) {
                    inputStream.reset();
                } else {
                    ((FileInputStream)inputStream).getChannel().position(0);
                }
            }

            if (total > Integer.MAX_VALUE) {
                // Later we cast to int, so check here.
                throw new IOException("file too large");
            }
            return (int)total;

        }
        // Cannot calculate length of input stream
        return -1;
    }

    private int transfer(StorageObject storageObject) throws IOException, GeneralSecurityException {

        final InputStream inputStream = storageObject.getInputStream();
        final boolean upload = inputStream != null;
        final Map<String,String> headers = new HashMap<String, String>();

        // TODO: content length of storage object shouldn't have to be calculated here
        if (upload) {
            long contentLength = storageObject.getLength();
            if (contentLength == -1) {
                contentLength = getContentLength(inputStream);
            }

            final String contentType = storageObject.getType() != null
                ? storageObject.getType()
                : "application/octet-stream";
            // put header fields names in lower-case as is expected by Object Storage signing
            headers.put("content-type", contentType);
            headers.put("content-length", Long.toString(contentLength));
        }

        int retries = 0;
        while (retries++ < RETRY_LIMIT) {

            final int responseCode = transferContent(storageObject, headers);

            switch (responseCode) {
                case HttpURLConnection.HTTP_OK:
                case HttpURLConnection.HTTP_CREATED:
                case HttpURLConnection.HTTP_PARTIAL:
                    return responseCode;
                case ERR_UPLOAD_CANCELLED:
                    getLogger().log(Level.INFO, (upload ? "Upload" : "Download") + " cancelled " + storageObject.getURI());
                    return upload ? HttpURLConnection.HTTP_CREATED : HttpURLConnection.HTTP_OK;
                case HttpURLConnection.HTTP_UNAUTHORIZED:
                    // retry with new token
                    storageAuthenticationData = null;
                    break;
                case ERR_BAD_CHECKSUM:
                case ERR_WRITING_REQUEST_BODY_TO_SERVER:
                    // retry
                    break;
                default:
                    return responseCode;
            }

            if (retries < RETRY_LIMIT) {
                if (inputStream.markSupported()) {
                    inputStream.reset();
                } else if (inputStream instanceof FileInputStream) {
                    FileInputStream fs = (FileInputStream)inputStream;
                    FileChannel fc = fs.getChannel();
                    fc.position(0);
                } else {
                    // cannot reset stream for retry
                    getLogger().log(Level.SEVERE, "Cannot reset input stream for retry");
                    return HttpURLConnection.HTTP_INTERNAL_ERROR;
                }
            }
        }
        getLogger().log(Level.INFO, "Retries exceeded " + (upload ? "uploading" : "downloading") + " " + storageObject.getURI());
        return HttpURLConnection.HTTP_INTERNAL_ERROR;
    }

    private int transferContent(StorageObject storageObject, Map<String,String> headers)
        throws IOException, GeneralSecurityException {

        final URL url;
        try {
            final String storageUri = storageObject.getURI();
            url = new URL(storageUri);
        } catch (MalformedURLException ex) {
            throw new IllegalArgumentException("Storage Cloud URL is invalid", ex);
        }

        final String encoding = storageObject.getEncoding();
        final InputStream inputStream = storageObject.getInputStream();
        final OutputStream outputStream = storageObject.getOutputStream();
        final boolean upload = inputStream != null;

        final String name = getObjectNameFromURL(url);
        final boolean isClassicObjectStorage = pathIsClassicObjectStore(url.getPath());
        final byte[] authenticationRequestBody
            = isClassicObjectStorage
            ? AUTHENTICATION_REQUEST_BODY_IS_NOT_NEEDED
            : createAuthenticationRequestBody(name, upload ? "PUT" : "GET");

        setStorageAuthenticationData(authenticationRequestBody);

        final HttpURLConnection connection = (HttpURLConnection) url.openConnection();
        if (upload) {
            connection.setDoOutput(true);
            connection.setRequestMethod("PUT");

            int contentLength;
            try {
                final String value = headers.get("content-length");
                contentLength = value != null ? Integer.valueOf(value) : -1;
            } catch (NumberFormatException e) {
                contentLength = -1;
            }
            if (contentLength == -1) {
                connection.setChunkedStreamingMode(CHUNK_SIZE);
            } else {
                connection.setFixedLengthStreamingMode(contentLength);
            }
            if (encoding != null) {
                connection.setRequestProperty("Content-Encoding", encoding);
            }
            Map<String, String> metadata = storageObject.getCustomMetadata();
            for (Map.Entry<String, String> entry : metadata.entrySet()) {
                String k = entry.getKey();
                String v = entry.getValue();
                if (isClassicObjectStorage) {
                    connection.setRequestProperty("X-Object-Meta-" + k, v);
                } else {
                    connection.setRequestProperty("opc-meta-" + k, v);
                }
            }
        } else {
            connection.setRequestMethod("GET");
            connection.setDoInput(true);
        }

        if (isClassicObjectStorage) {
            connection.setRequestProperty(AUTH_TOKEN_HEADER, storageAuthenticationData.getAuthToken());
        } else {
            for (Map.Entry<String,String> header : storageAuthenticationData.getHeaders().entrySet()) {
                connection.setRequestProperty(header.getKey(), header.getValue());
            }
        }

        for (Map.Entry<String,String> header : headers.entrySet()) {
            connection.setRequestProperty(header.getKey(), header.getValue());
        }

        connection.connect();

        int responseCode;
        final OutputStream os = upload ? connection.getOutputStream() : outputStream;
        final InputStream is = upload ? inputStream : connection.getInputStream();

        try {

            MessageDigest digest = null;
            try {
                digest = MessageDigest.getInstance("MD5");
            } catch (NoSuchAlgorithmException ex) {
                getLogger().log(Level.WARNING, "Storage Cloud Service: checksum could not be verified.", ex);
            }

            long transferredBytes = 0;
            int len;
            final byte[] b = new byte[CHUNK_SIZE];
            while ((len = is.read(b)) != -1) {
                if (((StorageObjectDelegate) storageObject).isCancelled()) {
                    return ERR_UPLOAD_CANCELLED;
                }
                os.write(b, 0, len);

                if (digest != null) {
                    digest.update(b, 0, len);
                }
                ((StorageObjectDelegate) storageObject)
                    .setTransferredBytes(transferredBytes += len);

            }
            final byte[] contentChecksum = digest.digest();

            responseCode = connection.getResponseCode();

            final String checksum =
                isClassicObjectStorage
                    ? connection.getHeaderField("ETag")
                    : connection.getHeaderField("opc-content-md5");

            final String date = connection.getHeaderField("Last-Modified");

            if (checksum != null && contentChecksum != null) {
                //Verify MD5 checksum (objects < 5GB)
                final String digestOut =
                    isClassicObjectStorage
                        ? bytesToHexString(contentChecksum)
                        : Base64.getEncoder().encodeToString(contentChecksum);
                if (!checksum.equals(digestOut)) {
                    getLogger().log(Level.INFO, "Storage Cloud Service: checksum mismatch");
                    return ERR_BAD_CHECKSUM;
                }
            }

            if (upload && (responseCode == HttpURLConnection.HTTP_CREATED || responseCode == HttpURLConnection.HTTP_OK)) {
                ((StorageObjectDelegate)storageObject).setAttributes(date, transferredBytes);
            } else if (!upload && (responseCode == HttpURLConnection.HTTP_OK || responseCode == HttpURLConnection.HTTP_PARTIAL)) {
                Map<String, List<String>> responseHeaders = connection.getHeaderFields();
                Set<Map.Entry<String, List<String>>> entrySet = responseHeaders.entrySet();
                for (Map.Entry<String, List<String>> entry : entrySet) {
                    try {
                        String key = entry.getKey();
                        List<String> values = entry.getValue();
                        // if there are no key or no value then drop it
                        if (key == null || values == null || values.size() == 0) {
                            continue;
                        }
                        if (key.startsWith("X-Object-Meta-")) {
                            storageObject.setCustomMetadata(key.substring("X-Object-Meta-".length()), values.get(0));
                        } else if (key.startsWith("opc-meta-")) {
                            storageObject.setCustomMetadata(key.substring("opc-meta-".length()), values.get(0));
                        }
                    } catch (IllegalArgumentException iae) {
                    } catch (NullPointerException ex) {
                    }
                }
            }

        } catch (IOException e) {
            if ("Error writing request body to server".equalsIgnoreCase(e.getMessage())) {
                getLogger().log(Level.SEVERE, e.toString());
                connection.disconnect();
                return ERR_WRITING_REQUEST_BODY_TO_SERVER;
            }
            throw e;
        } finally {

            // Close only the HttpURLConnection stream.
            // Closing StorageObject streams is responsibility of the caller to sync(StorageObject).
            if (upload) {
                try {
                    os.close();
                } catch (IOException ignored) {
                    // not expected to happen
                }
            } else {
                try {
                    is.close();
                } catch (IOException ignored) {
                    // not expected to happen
                }
            }
        }

        return responseCode;
    }

    private static boolean pathIsClassicObjectStore(String path) throws IllegalArgumentException {
        if (path.startsWith("/v1/")) {
            return true;
        }
        if (!path.startsWith("/n/")) {
            throw new IllegalArgumentException("path does not start with expected tokens");
        }
        return false;
    }

    // package to allow unit test
    static String getObjectNameFromURL(URL storageCloudUrl) throws IllegalArgumentException {
        Matcher matcher = OBJECT_STORAGE_PATTERN.matcher(storageCloudUrl.toString());
        if (matcher.matches()) {
            if (matcher.groupCount() > 0) {
                return matcher.group(1);
            }
        } else {
            matcher = OBJECT_STORAGE_CLASSIC_PATTERN.matcher(storageCloudUrl.toString());
            if (matcher.matches()) {
                if (matcher.groupCount() > 0) {
                    return matcher.group(1);
                }
            }
        }
        throw new IllegalArgumentException("path does not contain object name");
    }

    /**
     * Generate request body for POST /iot/api/v2/provisioner/storage.
     * See https://docs.cloud.oracle.com/iaas/Content/API/Concepts/signingrequests.htm.
     * @param objectName name for object in storage cloud, not {@code null}
     * @param method method being requested in storage cloud API, GET, HEAD, or POST, not {@code null}
     * @return request body for POST /iot/api/v2/provisioner/storage
     */
    private static byte[] createAuthenticationRequestBody(
        String objectName,
        String method) {

        assert objectName != null;
        assert method != null;

        byte[] requestBody = null;
        try {
            //
            // {
            //    "object":"object_name"
            //    "method":"PUT"
            // }
            //
            JSONObject authRequestBody = new JSONObject()
                .put("object", objectName)
                .put("method", method.toUpperCase(Locale.ROOT));
            requestBody = authRequestBody.toString().getBytes("UTF-8");
        } catch (JSONException e) {
            throw new RuntimeException(e.getMessage(), e);
        } catch (UnsupportedEncodingException e) {
            // cannot happen - UTF-8 is required encoding
            throw new RuntimeException(e.getMessage(), e);
        }
        return requestBody;
    }

    /**
     * Get the storage authentication from the IoT CS.
     * @param authenticationRequestBody the body of the authentication request.
     * Required for Object Storage, otherwise must be {@code null}.
     * @return the StorageAuthenticationResponse to the request
     * @throws IOException
     * @throws GeneralSecurityException 
     */
    private void setStorageAuthenticationData(byte[] authenticationRequestBody)
        throws IOException, GeneralSecurityException {

        // For Object Storage Classic, only get storage authentication if needed.
        // Note that for Object Storage Classic, storageAuthenticationData has to
        // be set to null whenever a response from the server is not
        // an OK or ACCEPTED.
        // For Object Storage, always get storage authentication.
        if (authenticationRequestBody == null && storageAuthenticationData != null) {
                return;
        }

        // Reset storageAuthenticationData to null so that it will remain null if
        // there is an exception from the POST/GET /iot/api/v2/provisioner/storage,
        // or response status is other than HTTP 200. This is to ensure against
        // the code using bad auth data.
        storageAuthenticationData = null;

        final HttpResponse response;
        if (authenticationRequestBody == null) {
            response = secureConnection.get(REST_STORAGE_AUTHENTICATION);
        } else {
            assert authenticationRequestBody != null;
            response = secureConnection.post(REST_STORAGE_AUTHENTICATION, authenticationRequestBody);
        }

        int status = response.getStatus();

        if (status != HttpURLConnection.HTTP_OK) {
            final String requestMethod = (authenticationRequestBody == null) ? "GET" : "POST";
            throw new TransportException(requestMethod, REST_STORAGE_AUTHENTICATION, response);
        }

        String jsonResponse = new String(response.getData(), "UTF-8");
        JSONObject json;
        try {
            json = new JSONObject(jsonResponse);
            if (getLogger().isLoggable(Level.FINER)) {
                getLogger().log(Level.FINER, json.toString(4));
            }
        } catch (JSONException e) {
            throw new RuntimeException(e);
        }

        storageAuthenticationData  = StorageAuthenticationResponse.fromJson(json);
    }

    private static String bytesToHexString(byte[] in) {
        final StringBuilder sb = new StringBuilder();
        for(byte b : in) {
            sb.append(String.format("%02x", b));
        }
        return sb.toString();
    }

    private static final Logger LOGGER = Logger.getLogger("oracle.iot.client.StorageConnection");
    private static Logger getLogger() { return LOGGER; }
}
