/*
 * Copyright (c) 2019, Oracle and/or its affiliates.  All rights reserved.
 *
 * This software is dual-licensed to you under the MIT License (MIT) and
 * the Universal Permissive License (UPL).  See the LICENSE file in the root
 * directory for license terms.  You may choose either license, or both.
 */

package com.oracle.iot.client.impl.http;

import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.GeneralSecurityException;

/**
 * Platform specific details for HttpClient
 */
class HttpClientBase {

    private static final int DEFAULT_RESPONSE_TIMEOUT = 15000; // milliseconds
    private static final int responseTimeout;
    static {
        int value = DEFAULT_RESPONSE_TIMEOUT;
        try {
            value = Integer.getInteger(
                "oracle.iot.client.http_response_timeout",
                DEFAULT_RESPONSE_TIMEOUT);
        } catch (SecurityException e) {
            // use default value
            value = DEFAULT_RESPONSE_TIMEOUT;
        } finally {
            responseTimeout = value > 0 ? value : 0;
        }
    }

    private final URL url;

    protected HttpClientBase(URL url) {
        if (url == null) {
            throw new IllegalArgumentException("url cannot be null");
        }
        this.url = assignStreamHandler(url);
    }

    /**
     * Get the HTTP response timeout.
     * The timeout The default value is 15000 milliseconds and can be set
     * with the {@code oracle.iot.client.http_response_timeout} property.
     * @return the HTTP response timeout
     */
    protected int getResponseTimeout() {
        return responseTimeout;
    }

    protected URL getUrl() {
        return url;
    }

    // Give implementation an opportunity to modify the connection before
    // the URLConnection {@code connect()} method is called.
    protected void beforeConnect(HttpURLConnection connection)
        throws IOException, GeneralSecurityException {
        // Override as needed
    }

    // If client is run in a weblogic environment, then the HTTPS implementation is
    // a weblogic.net.http.SOAPHttpsURLConnection, not a javax.net.ssl.HttpsURLConnection.
    // This is overcome by setting the URLStreamHandler on the URL to be the
    // default javase handler. See IOT-35530. See also IOT-40999.
    private static URL assignStreamHandler(URL url) {
        try {
            return new URL(
                null,
                url.toExternalForm(),
                new sun.net.www.protocol.https.Handler()
            );
        } catch (MalformedURLException ignored) {
            // This should not happen since the 'url' arg is well formed.
            return url;
        }
    }

}
